#!/usr/bin/env python

#############################################################################
##
# This file is part of Taurus
##
# http://taurus-scada.org
##
# Copyright 2011 CELLS / ALBA Synchrotron, Bellaterra, Spain
##
# Taurus is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
##
# Taurus is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
##
# You should have received a copy of the GNU Lesser General Public License
# along with Taurus.  If not, see <http://www.gnu.org/licenses/>.
##
#############################################################################


"""
This is an example of how to assign specific labels to arbitrary positions
of a given axis using AxisItem.setTicks().

(Pure Qt)
"""

import sys
from pyqtgraph.Qt import QtGui
import pyqtgraph as pg
import numpy as np


if __name__ == "__main__":

    app = QtGui.QApplication([])

    w = pg.PlotWidget()

    # define a list of position,label tuples
    ticks = [list(zip((1, 2, 3, 7, 8), ("a", "b", "c", "d", "e")))]
    xax = w.getAxis("bottom")
    xax.setTicks(ticks)
    w.plot(np.arange(9))

    w.show()

    sys.exit(app.exec_())
